# $Id: control.rb,v 1.44 2005/02/11 16:25:18 toki Exp $

require 'forwardable'
require 'rucy/rep'
require 'rucy/params'
require 'rucy/monitor'

module Rucy
  class Control
    extend Forwardable

    def initialize(rep=Repository::DEFAULT)
      @rep = rep
      @fork = @rep.fetch(:Fork)
      @loader = nil
      @factory = nil
      @pw_enc = @rep.fetch(:PasswordEncryptor)
      @err_mon = SetupErrorMonitor.new
      @logfiles_output = Array.new
      @restart_signal = nil
      @properties = nil
      @page_path = nil
      @server = nil
      @start_time = nil
      @restart_count = nil
      @restart_time = nil
      @AdminParams_class = @rep.fetch(:AdminParams_class)
      @ServerParams_class = @rep.fetch(:ServerParams_class)
      @LoggingParams_class = @rep.fetch(:LoggingParams_class)
      @AccessLogParams_class = @rep.fetch(:AccessLogParams_class)
      @AliasList_class = @rep.fetch(:AliasList_class)
      @DocumentList_class = @rep.fetch(:DocumentList_class)
      @FilterList_class = @rep.fetch(:FilterList_class)
    end

    def load_path=(new_path)
      @loader = @rep.fetch(:DocumentLoader, new_path)
      new_path
    end

    def store_path=(new_path)
      @properties = @rep.fetch(:Properties, new_path)
      new_path
    end

    attr_writer :page_path
    attr_writer :restart_signal

    def_delegator :@err_mon, :load_errors, :setup_load_errors
    def_delegator :@err_mon, :logging_errors, :setup_logging_errors
    def_delegator :@err_mon, :alias_errors, :setup_alias_errors
    def_delegator :@err_mon, :document_errors, :setup_document_errors
    def_delegator :@err_mon, :filter_errors, :setup_filter_errors

    # for backward compatibility
    def load_errors
      setup_load_errors.map{|err|
	[ err.mod_path,
	  err.exception
	]
      }
    end

    # for backward compatibility
    def logging_errors
      setup_logging_errors.map{|err|
	{ :logfile => err.log_path,
	  :exception =>err.exception
	}
      }
    end

    # for backward compatibility
    def alias_errors
      setup_alias_errors.map{|err|
	{ :alias_path => err.alias_path,
	  :orig_path => err.orig_path,
	  :virtual_host => err.virtual_host,
	  :exception => err.exception
	}
      }
    end

    # for backward compatibility
    def doc_errors
      setup_document_errors.map{|err|
	{ :document => err.document,
	  :arguments => err.arguments,
	  :mount_path => err.mount_path,
	  :mount_mask => err.mount_mask,
	  :virtual_host => err.virtual_host,
	  :exception => err.exception
	}
      }
    end

   # for backward compatibility
    def filter_errors
      setup_filter_errors.map{|err|
	{ :filter => err.filter,
	  :arguments => err.arguments,
	  :attach_path => err.attach_path,
	  :attach_mask => err.attach_mask,
	  :virtual_host => err.virtual_host,
	  :exception => err.exception
	}
      }
    end

    attr_reader :start_time
    attr_reader :restart_count
    attr_reader :restart_time

    def add_logfile(output)
      @logfiles_output.push(output)
      nil
    end

    def server_setup(server)
      server_params = self.server_params
      server_params.server_setup(self, @err_mon, server)
      @properties.privilege = server.privilege

      logging_params = self.logging_params
      logging_params.server_setup(self, @err_mon, server)

      access_log_params = self.access_log_params
      access_log_params.server_setup(self, @err_mon, server)

      @loader.load{|mod_path, exception|
	@err_mon.capture_load_error(exception, mod_path)
	server.err("error: not loaded #{mod_path.inspect}: #{exception}")
      }
      @factory = @loader.factory
      @factory.add_option(:pw_enc, @pw_enc)
      @factory.setup

      folder = FolderDocument.new
      server.mount(folder, '/')

      alias_list = self.aliases
      alias_list.server_setup(self, @err_mon, server)

      document_list = self.documents
      document_list.server_setup(self, @err_mon, server, @factory)

      filter_list = self.filters
      filter_list.server_setup(self, @err_mon, server, @factory)

      admin_params = self.admin_params
      admin_params.server_setup(self, @err_mon, server, @loader, @factory, @pw_enc, @page_path)

      @start_time = server.start_time
      @restart_count = server.restart_count
      @restart_time = server.restart_time
      @server = server
      server.close_hook{|s|
	logfiles_close
      }

      nil
    end

    def server_restart
      @restart_signal.notify_restart
      nil
    end

    def server_close
      @restart_signal.notify_close
      nil
    end

    def logfiles_close
      for output in @logfiles_output
	output.close
	@server.debug("close #{output.path.inspect}.")
      end
      nil
    end

    def admin_params
      params = @rep.fetch(:AdminParams)
      params.load(@properties)
      params
    end

    def set_admin_params(params)
      case (params)
      when @AdminParams_class
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def server_params
      params = @rep.fetch(:ServerParams, @fork)
      params.load(@properties)
      params
    end

    def set_server_params(params)
      case (params)
      when @ServerParams_class
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def logging_params
      params = @rep.fetch(:LoggingParams)
      params.load(@properties)
      params
    end

    def set_logging_params(params)
      case (params)
      when @LoggingParams_class
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def access_log_params
      params = @rep.fetch(:AccessLogParams)
      params.load(@properties)
      params
    end

    def set_access_log_params(params)
      case (params)
      when @AccessLogParams_class
	params.save(@properties)
      else
	raise "unknown parameter set: #{params.inspect}"
      end
      nil
    end

    def aliases
      list = @rep.fetch(:AliasList)
      list.load(@properties)
      list
    end

    def set_aliases(list)
      case (list)
      when @AliasList_class
	list.save(@properties)
      else
	raise "unknown parameter set: #{list.inspect}"
      end
      nil
    end

    def documents
      list = @rep.fetch(:DocumentList, @factory)
      list.load(@properties)
      list
    end

    def set_documents(list)
      case (list)
      when @DocumentList_class
	list.save(@properties)
      else
	raise "unknown parameter list: #{list.inspect}"
      end
      nil
    end

    def filters
      list = @rep.fetch(:FilterList, @factory)
      list.load(@properties)
      list
    end

    def set_filters(list)
      case (list)
      when @FilterList_class
	list.save(@properties)
      else
	raise "unknown parameter list: #{list.inspect}"
      end
      nil
    end
  end
end
